"""Wrapper functions around an LLM chain."""

# NOTE: moved to llama_index/llm_predictor/base.py
# NOTE: this is for backwards compatibility

from llama_index.llm_predictor.base import (  # noqa: F401
    LLMChain,
    LLMMetadata,
    LLMPredictor,
)
